package starlogo;

import javax.swing.*;
import java.awt.*;
import java.awt.peer.ComponentPeer;
import java.awt.peer.ContainerPeer;
import java.awt.event.*;
import java.util.*;
import java.applet.*;

/**
 * This class customizes Swing's repaint behavior. Swing's usual behavior when collecting "dirty"
 * rectangles to be repainted is to compute the union of all dirty rectangles for a Component.
 * This can be wasteful if, for example, only two small areas of a Component need repainting, and
 * they happen to be located in opposite corners of the Component's rectangle. With this class
 * installed, Swing will not combine dirty rectangles that do not overlap at all. This means
 * keeping track of multiple rectangles for each component.
 * <p>
 * The following call installs <code>MultirectRepaintManager</code> globally (for the current thread):
 * <pre>
 *    RepaintManager.setCurrentManager(new MultirectRepaintManager());
 * </pre>
 * <p>
 * Note: This class uses Hashtables of ArrayLists of Rectangles, and is not designed to be particularly
 * efficient. It is possible, though it seems unlikely, that the overhead could noticeably impact performance.
 * 
 * @author David Greenspan
 * @see javax.swing.RepaintManager
 * @see <a href="http://java.sun.com/products/jfc/tsc/articles/painting/index.html#mgr">The RepaintManager</a>
 */

public class MultirectRepaintManager extends RepaintManager {
    
    // new versions of RepaintManager's hashtables, to hold ArrayLists instead of Rectangles
    private Hashtable dirtyComponentsNew = new Hashtable();
    private Hashtable tmpDirtyComponentsNew = new Hashtable();
    
    public synchronized void addDirtyRegion(JComponent c, int x, int y, int w, int h) {
        if ((w <= 0) || (h <= 0) || (c == null)) return;
	if ((c.getWidth() <= 0) || (c.getHeight() <= 0)) return;
        
        if (extendDirtyRegion(c, x, y, w, h)) {
	    return;
	} 
        
        /* Make sure that c and all it ancestors (up to an Applet or
         * Window) are visible.  This loop has the same effect as 
         * checking c.isShowing() (and note that it's still possible 
         * that c is completely obscured by an opaque ancestor in 
         * the specified rectangle).
         */
        Component root = null;

        // Note: We can't synchronize around this, Frame.getExtendedState
        // is synchronized so that if we were to synchronize around this
        // it could lead to the possibility of getting locks out
        // of order and deadlocking.
	for (Container p = c; p != null; p = p.getParent()) {
            if (!p.isVisible() || (p.getPeer() == null)) {
                return;
            }
            if ((p instanceof Window) || (p instanceof Applet)) {
                // Iconified frames are still visible!
                if (p instanceof Frame &&
		    (((Frame)p).getExtendedState() & Frame.ICONIFIED) ==
                                    Frame.ICONIFIED) {
                    return;
                }
                root = p;
                break;
            }
        }
	
        if (root == null) return;

        synchronized(this) {
            if (extendDirtyRegion(c, x, y, w, h)) {
                // In between last check and this check another thread
                // queued up runnable, can bail here.
                return;
            }

	    // associate a new ArrayList with this component
	    ArrayList v = new ArrayList();
	    v.add(new Rectangle(x,y,w,h));
	    dirtyComponentsNew.put(c,v);
            //dirtyComponents.put(c, new Rectangle(x, y, w, h));
        }

        // The super method uses some calls that aren't accessible to this package to queue a repaint.
        // Also, allowing the superclass to maintain its own list of rects allows us to leave alone
        // methods such as getDirtyRegion() and isCompletelyDirty()
        super.addDirtyRegion(c,x,y,w,h);
    }

    private synchronized boolean extendDirtyRegion(Component c, int x, int y, int w, int h) {
	ArrayList rects = (ArrayList)dirtyComponentsNew.get(c);
        if (rects != null) {
            // already have a rect vector for this component
            addRect(rects,x,y,w,h);
            return true;
        }
	return false;
    }
    
    public void paintDirtyRegions() {
        JComponent dirtyComponent;
        
	synchronized(this) {
            // RepaintManager.java says "swap for thread safety"
            Hashtable tmp = tmpDirtyComponentsNew;
	    tmpDirtyComponentsNew = dirtyComponentsNew;
	    dirtyComponentsNew = tmp;
	    dirtyComponentsNew.clear();
	}

        int count = tmpDirtyComponentsNew.size();
        if (count == 0) { 
	    return;
	}

        int localBoundsX = 0;
        int localBoundsY = 0;
        int localBoundsH = 0;
        int localBoundsW = 0;	
        Iterator keys;

        ArrayList roots = new ArrayList(count);
        keys = tmpDirtyComponentsNew.keySet().iterator();
        
        while(keys.hasNext()) {
            dirtyComponent = (JComponent)keys.next();
	    // the only way to tell RepaintManager to clear its list of rects
            super.markCompletelyClean(dirtyComponent); 
            collectDirtyComponents(tmpDirtyComponentsNew, dirtyComponent, roots);
        }

        for(Iterator i = roots.iterator(); i.hasNext(); ) {
            dirtyComponent = (JComponent)i.next();
            ArrayList rects = (ArrayList)tmpDirtyComponentsNew.get(dirtyComponent);
            
            localBoundsH = dirtyComponent.getHeight();
            localBoundsW = dirtyComponent.getWidth();
            
            intersectRect(rects,localBoundsX,localBoundsY,localBoundsW,localBoundsH);
            
            if (fullSizeRects(rects, dirtyComponent.getWidth(),
			      dirtyComponent.getHeight())) {
		Container parent = dirtyComponent.getParent();
		ComponentPeer parentPeer;
                if (parent != null && !parent.isLightweight() &&
                    (parentPeer = parent.getPeer()) != null) {
                    // Cancel any pending paints on the heavy weight peer.
                    // This avoid duplicate painting.
//                     ((ContainerPeer)parentPeer).cancelPendingPaint(
//                                     dirtyComponent.getX(),
//                                     dirtyComponent.getY(),
// 				    dirtyComponent.getWidth(),
//                                     dirtyComponent.getHeight());
                }
            }

            for(Iterator j = rects.iterator(); j.hasNext();) {
                Rectangle r = (Rectangle)j.next();
		//System.out.println("redraw rect: " + r);
                dirtyComponent.paintImmediately(r);
            }
        }
        
	tmpDirtyComponentsNew.clear();
    }


    protected void collectDirtyComponents(Hashtable dirtyComponentsNew, JComponent dirtyComponent, ArrayList roots) {
        int dx, dy, rootDx, rootDy;
        Component component, rootDirtyComponent, parent;
        Rectangle cBounds;

        component = rootDirtyComponent = dirtyComponent;

        int x = dirtyComponent.getX();
        int y = dirtyComponent.getY();
        int w = dirtyComponent.getWidth();
        int h = dirtyComponent.getHeight();

        //cBounds = dirtyComponent.getBounds();

        dx = rootDx = 0;
        dy = rootDy = 0;
        ArrayList tmp = copyRects((ArrayList)dirtyComponentsNew.get(dirtyComponent));
        
        intersectRect(tmp,0,0,w,h);

        if (tmp.size() < 1) return;

        for(;;) {
            parent = component.getParent();
            if (parent == null) break;

            if (!(parent instanceof JComponent)) break;

            component = parent;

            dx += x;
            dy += y;
            offsetRects(tmp,x,y);
            
            x = component.getX();
            y = component.getY();
            w = component.getWidth();
            h = component.getHeight();
            //cBounds = ((JComponent)component).getBounds();
            intersectRect(tmp,0,0,w,h);

            if (tmp.size() < 1) return;
            
            if (dirtyComponentsNew.get(component) != null) {
                rootDirtyComponent = component;
                rootDx = dx;
                rootDy = dy;
            }
        } 
        
        if (dirtyComponent != rootDirtyComponent) {
            offsetRects(tmp,rootDx - dx,rootDy - dy);
	    ArrayList v = (ArrayList)dirtyComponentsNew.get(rootDirtyComponent);
            addRects(v,tmp);
        }

        if (!roots.contains(rootDirtyComponent)) {
            roots.add(rootDirtyComponent);	
	}
    }

    public void markCompletelyClean(JComponent aComponent) {
        synchronized(this) {
                dirtyComponentsNew.remove(aComponent);
		super.markCompletelyClean(aComponent);
        }
    }

    
    ///// private utility functions
    
    private void addRect(ArrayList v,int x,int y,int w,int h) {
        Rectangle newRect = new Rectangle(x,y,w,h);
        for(ListIterator i = v.listIterator(v.size()); i.hasPrevious();) {
            Rectangle r = (Rectangle)i.previous();
            if (doesOverlap(x,y,w,h,r)) {
                i.remove();
                SwingUtilities.computeUnion(r.x,r.y,r.width,r.height,newRect);
            }
        }
        v.add(newRect);
    }
    
    private void addRects(ArrayList v, ArrayList rects) {
        for(Iterator i = rects.iterator(); i.hasNext();){ 
            Rectangle r = (Rectangle)i.next();
            addRect(v, r.x, r.y, r.width, r.height);
        }
    }
    
    private ArrayList copyRects(ArrayList v) {
        ArrayList newV = (ArrayList)v.clone();
        for(ListIterator i = newV.listIterator(); i.hasNext();) { 
            i.set(new Rectangle((Rectangle)i.next()));
        }
        return newV;
    }
    
    private void intersectRect(ArrayList v,int x,int y,int w,int h) {
        for(Iterator i = v.iterator(); i.hasNext();){ 
            SwingUtilities.computeIntersection(x,y,w,h,(Rectangle)i.next());
        }
        cleanRects(v);
    }
    
    private void cleanRects(ArrayList v) {
        for(ListIterator i = v.listIterator(v.size()); i.hasPrevious();) {
	    if (((Rectangle)i.previous()).isEmpty()) {
		i.remove();
	    }
        }
    }
    
    private void offsetRects(ArrayList v,int x,int y) {
        for(Iterator i = v.iterator(); i.hasNext();) {
            Rectangle r = (Rectangle)i.next();
            r.setLocation(r.x+x,r.y+y);
        }
    }
    
    private boolean fullSizeRects(ArrayList v, int w, int h) {
        for(Iterator i = v.iterator(); i.hasNext();) {
            Rectangle r = (Rectangle)i.next();
	    if (r.x == 0 && r.y == 0 &&
		r.width == w && r.height == h) {
		return true;
	    }
	}
	return false;
    }

    private boolean doesOverlap(int x,int y,int w,int h,Rectangle r) {
        return !(r.x+r.width < x || r.x > x+w || r.y+r.height < y || r.y > y+h);
    }
    
}
