/******************************************************************
OpenStarLogo

Originally developed by Mitchel Resnick, Andrew Begel, Eric Klopfer, Michael
Bolin, Corey McCaffrey, Molly Jones, Matthew Notowidigdo, Sebastian Ortiz, David
Greenspan, Michael Mandel, Tim Garnett, Max Goldman, Julie Kane, Russell
Zahniser, Weifang Sun, and Robert Tau at the MIT Teacher Education Program and
Media Laboratory, MIT, Cambridge, Massachusetts, with support from the National
Science Foundation and the LEGO Group. With previous work by Vanessa Colella,
Bill Thies, Brian Silverman, Monica Linden, Alice Yang, Ankur Mehta, Adam Eames,
Max Planck, and Sumita Kumar.

Copyright (c) 2006 MIT Teacher Education Program

Created by the MIT Teacher Education Program (http://education.mit.edu), and MIT
Media Lab (http://www.media.mit.edu) at the Massachusetts Institute of
Technology, Cambridge, MA. Permission is hereby granted, free of charge, to any
person obtaining a copy of this software and associated documentation files (the
"Software"), to deal in the Software without restriction, including without
limitation the rights to use, copy, modify, merge, publish, distribute,
sublicense, and/or sell copies of the Software, and to permit persons to whom
the Software is furnished to do so, subject to the following conditions:

1. The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software. 2. For rights to publish,
distribute, sublicense or sell the Software or its derivatives for commercial
purposes, please see the Technology Licensing Office, MIT, Five Cambridge
Center, Kendall Square Room NE25-230, Cambridge, MA 02142-1493, (617) 253-6966

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.


 *******************************************************************/

package starlogo;

import java.awt.Graphics2D;
import java.awt.print.*;
import javax.swing.SwingUtilities;
import java.awt.font.*;
import java.awt.geom.*;
import java.awt.Font;
import java.text.*;
import java.util.ArrayList;
import java.io.LineNumberReader;
import java.io.StringReader;
import java.io.IOException;

public class Print {

    PrinterJob printjob;
    static PageFormat pageformat;

    Printable frameToPrint;

    static void setupPage() {
	PrinterJob pj = PrinterJob.getPrinterJob();
	pageformat = pj.pageDialog(pj.defaultPage());
    }

    public Print(Printable caller, String title) {
	frameToPrint = caller;
	printjob = PrinterJob.getPrinterJob();
	//printjob.setName(title);
	if (pageformat == null) {
	    pageformat = printjob.defaultPage();
	}
	printjob.setPrintable(frameToPrint, pageformat);
    }

    public void print() {
	if (printjob.printDialog()) {
	    SwingUtilities.invokeLater(new Runnable() {
		    public void run() {
			try {
			    printjob.print();
			} catch (PrinterException e) {
			    // XXX Report error
			    e.printStackTrace();
			}
		    }
		});
	}
    }


    public static void scaleGraphics(Graphics2D g2d,
				     double pageWidth, double pageHeight,
				     double imageWidth, double imageHeight) {

	boolean tall = (imageWidth < imageHeight);

	if (imageWidth > pageWidth) {
	    if (imageHeight > pageHeight) {
		if (tall) {
		    double yscale = pageHeight / imageHeight;
		    double newWidth = pageWidth * yscale;
		    if (newWidth > pageWidth) {
			g2d.scale(pageWidth / imageWidth, yscale);
		    } else {
			g2d.scale(yscale, yscale);
		    }
		} else {
		    double xscale = pageWidth / imageWidth;
		    double newHeight = pageHeight * xscale;
		    if (newHeight > pageHeight) {
			g2d.scale(xscale, pageHeight / imageHeight);
		    } else {
			g2d.scale(xscale, xscale);
		    }
		}
	    } else {
		double scale = pageWidth / imageWidth;
		g2d.scale(scale, scale);
	    }
	} else if (imageHeight > pageHeight) {
	    double scale = pageHeight / imageHeight;
	    g2d.scale(scale, scale);
	}
    }

    
    public static class Printed {
	double endPoints;
	int pagesToGo;
	Printed(double endPoints, int pagesToGo) {
	    this.endPoints = endPoints;
	    this.pagesToGo = pagesToGo;
	}
    }

    public static Printed printText(Graphics2D g2d, String origText, double startPoints, 
			     int pageIndex, Font font,
			     double maxWidth, double maxHeight) {
	LineNumberReader lnr = new LineNumberReader(new StringReader(origText));
	while(pageIndex >= 0) {
	    String text = null;
	    try {
		text = lnr.readLine();
	    } catch (IOException e) {
		break;
	    }
	    if (text == null) break;   // we're done with the text
	    do { 
		LayoutHeight lh = printParagraph((pageIndex == 0) ? g2d : null, text, 
						 startPoints, maxWidth, maxHeight, font);
		if (lh.points > 0) {
		    startPoints = lh.points;
		    text = "";
		} else {
		    // the whole thing didn't fit. Split it across pages.
		    int caretPos = lh.caretPos; // caretPos chars did fit.
		    if (caretPos > 0) {
			text = text.substring(caretPos, text.length());
			startPoints = lh.points;
		    } else {
			// we were over the limit but didn't know it on the prev line.
			// Push it to the limit and go to the next
			// page.
			startPoints = maxHeight;
		    }
		}
		if (startPoints >= maxHeight) {
		    pageIndex--;
		    startPoints = 0;
		}
	    } while(text.length() > 0);
	    //System.out.println("start points after line is " + startPoints);
	    
	    if (startPoints >= maxHeight) {
		pageIndex--;
		startPoints = 0;
	    }
	}
	return new Printed(startPoints, pageIndex);
    }


    /*
     * return value for printParagraph 
     * if points < 0, then caretPos is the number of chars printed 
     * if points >= 0, then points is the ending y position in points
     * where the text finished printing 
     *
     */
    public static class LayoutHeight {
	int caretPos;
	double points;
	LayoutHeight(int caretPos, double points) {
	    this.caretPos = caretPos;
	    this.points = points;
	}
    }


    public static LayoutHeight printParagraph(Graphics2D g2d, String text, double startPoints, 
					      double maxWidth, double maxHeight,
					      Font font) {

	FontRenderContext frc = new FontRenderContext(null, true, true);

	// printing an empty line? skip some space.
	if (text.length() == 0) {
	    LineMetrics lm = font.getLineMetrics("WWW" /* dummy text */, frc);
	    return new LayoutHeight(0,
				    startPoints + lm.getAscent() + 
				    lm.getDescent() + lm.getLeading());
	}
				    
	AttributedString paragraphText = new AttributedString (text);
	//--- Set the font for this text
	paragraphText.addAttribute (TextAttribute.FONT, font);
	
	//--- Create a point object to set the top left corner of the TextLayout object
	Point2D.Double pen = new Point2D.Double (0, startPoints);

         //--- Set the width of the TextLayout box
	double width = maxWidth;

	//--- Create a LineBreakMeasurer to wrap the text for the TextLayout object
	//--- Note the second parameter, the FontRendereContext. I have set the second
	//--- parameter antiAlised to true and the third parameter useFractionalMetrics
	//--- to true to get the best possible output
	LineBreakMeasurer lineBreaker = 
	    new LineBreakMeasurer (paragraphText.getIterator(), frc);
	
	//--- Create the TextLayouts object
	TextLayout layout;
	
	//--- Create a ArrayList to temporarily store each line of text
	ArrayList lines = new ArrayList ();
	
	//--- Get the output of the LineBreakMeasurer and store it in a ArrayList
	while ((layout = lineBreaker.nextLayout ((float) width)) != null) {
            lines.add (layout);
	}
	int chars = 0;
	//--- Scan each line of the paragraph and print it
	for (int i = 0; i < lines.size (); i++) {
	    //--- Get the line from the vector
	    layout = (TextLayout) lines.get (i);
	    
	    //--- Align the Y pen to the ascend of the font, remember that
	    //--- the ascend is origin (0, 0) of a font. Refer to figure 1
	    pen.y += layout.getAscent ();
	    

	    // will the next line go over the limit? if so, don't draw it.
	    if ((pen.y + layout.getDescent() + layout.getLeading()) > maxHeight) {
		return new LayoutHeight(chars, -1);
	    }
	    
	    //--- Draw the line of text
	    if (g2d != null) layout.draw (g2d, (float) pen.x, (float) pen.y);
	    
	    //--- Move the pen to the next position adding the descent and
	    //--- the leading of the font
	    pen.y += layout.getDescent () + layout.getLeading ();

	    // keep a character count of all the chars we've drawn
	    chars += layout.getCharacterCount();
	}
	return new LayoutHeight(chars, pen.y);
    }

}


