/******************************************************************
OpenStarLogo

Originally developed by Mitchel Resnick, Andrew Begel, Eric Klopfer, Michael
Bolin, Corey McCaffrey, Molly Jones, Matthew Notowidigdo, Sebastian Ortiz, David
Greenspan, Michael Mandel, Tim Garnett, Max Goldman, Julie Kane, Russell
Zahniser, Weifang Sun, and Robert Tau at the MIT Teacher Education Program and
Media Laboratory, MIT, Cambridge, Massachusetts, with support from the National
Science Foundation and the LEGO Group. With previous work by Vanessa Colella,
Bill Thies, Brian Silverman, Monica Linden, Alice Yang, Ankur Mehta, Adam Eames,
Max Planck, and Sumita Kumar.

Copyright (c) 2006 MIT Teacher Education Program

Created by the MIT Teacher Education Program (http://education.mit.edu), and MIT
Media Lab (http://www.media.mit.edu) at the Massachusetts Institute of
Technology, Cambridge, MA. Permission is hereby granted, free of charge, to any
person obtaining a copy of this software and associated documentation files (the
"Software"), to deal in the Software without restriction, including without
limitation the rights to use, copy, modify, merge, publish, distribute,
sublicense, and/or sell copies of the Software, and to permit persons to whom
the Software is furnished to do so, subject to the following conditions:

1. The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software. 2. For rights to publish,
distribute, sublicense or sell the Software or its derivatives for commercial
purposes, please see the Technology Licensing Office, MIT, Five Cambridge
Center, Kendall Square Room NE25-230, Cambridge, MA 02142-1493, (617) 253-6966

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.


 *******************************************************************/

package starlogo;

import java.io.*;
import java.util.Hashtable;
import yoyo.*;
import java.awt.image.MemoryImageSource;
import java.awt.Image;
import java.awt.Toolkit;
import java.net.*;

public class TurtleShapes {

    public static final int MAXSHAPES = 64;
	
    public static final int ON = 0xFFFFFFFF;
    public static final int OFF = 0x00000000;
	
    public static final int SMALL = 8;
    public static final int MEDIUM = 16;
    public static final int LARGE = 32;

    static TurtleShapes[] shapes = new TurtleShapes[MAXSHAPES];
	
    static int numShapes = 0;

    static ShapeClipboard clipboard = null;

    /*********************************************************/
	
    String name;
    int id;
    boolean rotatep;
    double[][] colors8x8 = new double[16][];
    int[][] masks8x8     = new int[16][];
    double[][] colors16x16 = new double[16][];
    int[][] masks16x16     = new int[16][];
    double[][] colors32x32 = new double[16][];
    int[][] masks32x32     = new int[16][];

    // colors array is:
    // SMALL  64   (8x8)
    // MEDIUM 256  (16x16)
    // LARGE  1024 (32x32)
    // masks array is twice as large as colors array
    // mask_index mod 2 = 0 turtle transparent
    // mask_index mod 2 = 1 patch transparent
    // masks for colors[i] is masks[2*i] and masks[2*i+1]

    TurtleShapes(String name, int id, 
		 double[] colors8x8, int[] masks8x8,
		 double[] colors16x16, int[] masks16x16,
		 double[] colors32x32, int[] masks32x32,
		 boolean rotatep) {
	
	this.name = name;
	this.id = id;
	this.rotatep = rotatep;

	this.colors8x8[0] = colors8x8;
	this.colors16x16[0] = colors16x16;
	this.colors32x32[0] = colors32x32;
 
	this.masks8x8[0] = masks8x8;
	this.masks16x16[0] = masks16x16;
	this.masks32x32[0] = masks32x32;

	reRotateShapes();
    }

    void reRotateShapes() {
	if (rotatep) {
	    for (int i=1; i<16; i++) {
		if (colors8x8[i] == null || colors8x8[i] == colors8x8[0]) 
		    colors8x8[i] = new double[SMALL*SMALL];
		if (masks8x8[i]  == null || masks8x8[i] == masks8x8[0]) 
		    masks8x8[i]  = new int[2*SMALL*SMALL];
		rotate(i, colors8x8[0], masks8x8[0], colors8x8[i], masks8x8[i]);
		
		if (colors16x16[i] == null || colors16x16[i] == colors16x16[0]) 
		    colors16x16[i] = new double[MEDIUM*MEDIUM];
		if (masks16x16[i]  == null || masks16x16[i] == masks16x16[0]) 
		    masks16x16[i]  = new int[2*MEDIUM*MEDIUM];
		rotate(i, colors16x16[0], masks16x16[0], colors16x16[i], masks16x16[i]);
		
		if (colors32x32[i] == null || colors32x32[i] == colors32x32[0]) 
		    colors32x32[i] = new double[LARGE*LARGE];
		if (masks32x32[i]  == null || masks32x32[i] == masks32x32[0])
		    masks32x32[i]  = new int[2*LARGE*LARGE];
		rotate(i, colors32x32[0], masks32x32[0], colors32x32[i], masks32x32[i]);
	    }
	} else { 
	    for (int i = 0; i < 16; i++) {
		colors8x8[i] = colors8x8[0];
		masks8x8[i]  = masks8x8[0];
		colors16x16[i] = colors16x16[0];
		masks16x16[i]  = masks16x16[0];
		colors32x32[i] = colors32x32[0];
		masks32x32[i]  = masks32x32[0];
	    }
	}
    }

    double[][] getColorsForSize(int size) {
	switch (size) {
	case SMALL: return colors8x8;
	case MEDIUM: return colors16x16;
	case LARGE: return colors32x32;
	} 
	return null;
    }

    int[][] getMasksForSize(int size) {
	switch (size) {
	case SMALL: return masks8x8;
	case MEDIUM: return masks16x16;
	case LARGE: return masks32x32;
	} 
	return null;
    }

    void setColorsForSize(double[] colors, int size) {
	switch (size) {
	case SMALL: colors8x8[0] = colors; break;
	case MEDIUM: colors16x16[0] = colors; break;
	case LARGE: colors32x32[0] = colors; break;
	} 
    }

    void setMasksForSize(int[] masks, int size) {
	switch (size) {
	case SMALL: masks8x8[0] = masks; break;
	case MEDIUM: masks16x16[0] = masks; break;
	case LARGE: masks32x32[0] = masks; break;
	} 
    }

    

    Image getShapeAsImage(Turtle t, int size) {
	double[][] colors = getColorsForSize(size);
	int[][] masks = getMasksForSize(size);
	int numpixels = colors[0].length;
	int[] pixels = new int[numpixels * 16];
	int shapestart = 0;
	int turtlec = (int)(t.color * 32.0);
	int tc = Colors.colorarrayints[turtlec];
	for(int h = 0; h < 16; h++) {
	    //System.out.println("heading = " + h);
	    int linestart = shapestart;
	    for(int y = 0; y < size; y++) {
		for (int x = 0; x < size; x++) {
		    int index = x + y * size;
		    int index2 = 2 * index;
		    int tmask = masks[h][index2];
		    int pmask = masks[h][index2+1];
		    int c = Colors.colorarrayints[(int)(colors[h][index] * 32.0)];
		    pixels[linestart + x] = (tmask & tc) | c;
		    if (pmask == ON) pixels[linestart + x] = 0;
		    //System.out.println("pixels[" + (x + y*size + numpixels*h) + "] = 0x" + Integer.toHexString(pixels[x + y*size + numpixels*h]));
		    
		}
		linestart += size * 16;
	    }
	    shapestart += size;
	}
	return Toolkit.getDefaultToolkit().createImage(new MemoryImageSource(size * 16, size, Colors.tcm, pixels, 0, size * 16));
    }

    static void clearHash() {
	if (shapes == null) return;
	for (int i = 0; i < shapes.length; i++) {
	    if (shapes[i] != null && shapes[i].shapeImages != null) {
		shapes[i].shapeImages.clear();
	    }
	}
    }

    Hashtable shapeImages = new Hashtable(100);

    static int shapeHash(int color, int size) {
	size = (((size>>3)-1) & 0x3) << 16;
	color = color & 0x1FFF;
	return size | color;
    }

    static Image getShapeAsImage(Turtle t, int cellsize, Application app) {
	TurtleShapes ts = getShape(t.shape);
	int heading = (int) (((t.heading+11.25) / 22.5) % 16);
	int tc = (int)(t.color * 32.0);
	Integer hash = new Integer(shapeHash(tc, cellsize));
	//System.out.println("heading = " + heading + " turtlec = " + tc + " shape hash: " + hash);
	Image im = (Image)ts.shapeImages.get(hash);
	if (im != null) return im;
	im = ts.getShapeAsImage(t, cellsize);
	ts.shapeImages.put(hash, im);
	return im;
    }
    

    static void loadDefaultShapes(Application app) {
	try {
	    //this should be handled specially along with image files
	    String name = app.textBundle.getString("shapes_file");
	    //System.out.println("default shape name file is " + name);
		URL  u =  TurtleShapes.class.getResource(name);
		//System.out.println("default shape name url is " + u.toString());
		String defaultShapeString = (String)PrimFile.prim_urltostring(u, app.sl.c);
		loadShapes(defaultShapeString, true, app);
	} catch (Exception e) {
	    app.err.println("Default shapes didn't load properly.");
	    e.printStackTrace(app.err);
	    // temporary -- replace this by loading them from a fil called defaultshapes.txt
	    //shapes8x8[0] = new TurtleShapes("turtle", 0, createBlankColors(8), true);
	    setShapeToSquareShape(1, app);
	    //shapes16x16[0] = new TurtleShapes("turtle", 0, createBlankColors(16), true);
	    setShapeToTurtleShape(0, app);
	    //shapes32x32[0] = new TurtleShapes("turtle", 0, createBlankColors(32), true);
	    setShapeToPersonShape(2, app);
	    for(int i = 3; i < MAXSHAPES; i++) {
		setShapeToBlankShape(i, app);
		shapes[i].name = app.textBundle.getString("shape_label") + i;
	    }
	    numShapes = MAXSHAPES;
	}
    }

    static void setName(String name, int number) {
	if (name == null) name = "";
	if (shapes[number] != null) shapes[number].name = name;
    }

    static boolean conflictingShapeName(String name, int notme) {
	for(int i = 0; i < MAXSHAPES; i++) {
	    if (i == notme) continue;
	    if (shapes[i] != null) {
		if (name.equalsIgnoreCase(shapes[i].name)) {
		    return true;
		}
	    } 
	}
	return false;
    }
    
    static String[] getShapeNames() {
	String[] names = new String[MAXSHAPES];
	for(int i = 0; i < MAXSHAPES; i++) {
	    if (shapes[i] != null) {
		names[i] = shapes[i].name;
	    } else {
		names[i] = "";
	    }
	}
	return names;
    }

    static int calculateNumShapes() {
	int num = 0;
	for (int i = 0; i < shapes.length; i++) {
	    if (shapes[i] != null) num++;
	}
	return num;
    }


    static TurtleShapes getShape(int shapenum) {
	if (shapenum < 0 || shapenum >= MAXSHAPES) return null;
	return shapes[shapenum];
    }

    static double[] copyColors(double[] colors) {
	double[] newcolors = new double[colors.length];
	System.arraycopy(colors, 0, newcolors, 0, colors.length);
	return newcolors;
    }

    static int[] copyMasks(int[] masks) {
	int[] newmasks = new int[masks.length];
	System.arraycopy(masks, 0, newmasks, 0, masks.length);
	return newmasks;
    }

    /**
     * @param c true if copying turtle shape, false if cutting.
     */
    TurtleShapes copy() {

	double[] cols8x8 = copyColors(colors8x8[0]);
	int[] msks8x8 = copyMasks(masks8x8[0]);

	double[] cols16x16 = copyColors(colors16x16[0]);
	int[] msks16x16 = copyMasks(masks16x16[0]);

	double[] cols32x32 = copyColors(colors32x32[0]);
	int[] msks32x32 = copyMasks(masks32x32[0]);

	return new TurtleShapes(name, -1, cols8x8, msks8x8, cols16x16, msks16x16,
				cols32x32, msks32x32, rotatep);

    }

    static void cutShape(int i, Application app) {
	clipboard = new ShapeClipboard(i, true);
	setShapeToBlankShape(i, app);
    }

    static void copyShape(int i, Application app) {
	clipboard = new ShapeClipboard(i, false);
    }

    static void pasteShape(int i, Application app) {
	if (clipboard == null) return;
	shapes[i]      = clipboard.shape.copy();
	if (!clipboard.cutp) {
	    int j = 2;
	    String originalName = shapes[i].name + app.textBundle.getString("_copy");
	    shapes[i].name = originalName;
	    while(conflictingShapeName(shapes[i].name, i)) {
		shapes[i].name = originalName + "-" + (j++);
	    }
	}
	shapes[i].id   = i;
    }


    static void setShapeToBlankShape(int shape, Application app) {
	int[] masks8x8 = createBlankMasks(SMALL);
	double[] colors8x8 = createBlankColors(SMALL);

	int[] masks16x16 = createBlankMasks(MEDIUM);
	double[] colors16x16 = createBlankColors(MEDIUM);

	int[] masks32x32 = createBlankMasks(LARGE);
	double[] colors32x32 = createBlankColors(LARGE);
	
	String name = app.textBundle.getString("unknown_shape");
	shapes[shape] = new TurtleShapes(name,
					 shape, colors8x8, masks8x8,
					 colors16x16, masks16x16, 
					 colors32x32, masks32x32, false);
    }

    static void setShapeToSquareShape(int shape, Application app) {
	int[] masks8x8 = new int[2 * SMALL * SMALL];
	for (int i = 0; i<masks8x8.length; i+=2) { 
	    masks8x8[i] = ON;
	    masks8x8[i+1] = OFF;
	}

	double[] colors8x8 = createBlankColors(SMALL);
	
	double[] colors16x16 = new double[MEDIUM*MEDIUM];
	int[] masks16x16 = new int[2 * MEDIUM * MEDIUM];
	resize(colors8x8, masks8x8, MEDIUM, colors16x16, masks16x16, app);

	double[] colors32x32 = new double[LARGE*LARGE];
	int[] masks32x32 = new int[2 * LARGE * LARGE];
	resize(colors8x8, masks8x8, LARGE, colors32x32, masks32x32, app);

	String name = app.textBundle.getString("square_shape");
	shapes[shape] = new TurtleShapes(name, 
					 shape, colors8x8, masks8x8,
					 colors16x16, masks16x16, 
					 colors32x32, masks32x32, false);
    }


    static void setShapeToTurtleShape(int shape, Application app) {

	double[] colors = new double[MEDIUM * MEDIUM];
	int[] masks     = new int[2 * MEDIUM * MEDIUM];
	makeTurtleCase(colors, masks);

	double[] colors8x8 = new double[SMALL*SMALL];
	int[] masks8x8 = new int[2 * SMALL * SMALL];
	resize(colors, masks, SMALL, colors8x8, masks8x8, app);

	double[] colors32x32 = new double[LARGE*LARGE];
	int[] masks32x32 = new int[2 * LARGE * LARGE];
	resize(colors, masks, LARGE, colors32x32, masks32x32, app);

	String name = app.textBundle.getString("turtle_shape");
	shapes[shape] = new TurtleShapes(name, 
					 shape, colors8x8, masks8x8, 
					 colors, masks, 
					 colors32x32, masks32x32, true);
    }

    static void setShapeToPersonShape(int shape, Application app) {

	double[] colors = new double[MEDIUM * MEDIUM];
	int[] masks     = new int[2 * MEDIUM * MEDIUM];
	makePersonCase(colors, masks);

	double[] colors8x8 = new double[SMALL*SMALL];
	int[] masks8x8 = new int[2 * SMALL * SMALL];
	resize(colors, masks, SMALL, colors8x8, masks8x8, app);

	double[] colors32x32 = new double[LARGE*LARGE];
	int[] masks32x32 = new int[2 * LARGE * LARGE];
	resize(colors, masks, LARGE, colors32x32, masks32x32, app);

	String name = app.textBundle.getString("person_shape");
	shapes[shape] = new TurtleShapes(name, 
					 shape, colors8x8, masks8x8, 
					 colors, masks, 
					 colors32x32, masks32x32, true);
    }


    static double[] createBlankColors(int sizetocreate) {
	double[] colors = new double[sizetocreate * sizetocreate];
	for (int i = 0; i < colors.length; i += 3) {
	    colors[i] = OFF;
	}
	return colors;
    }
    
    static int[] createBlankMasks(int sizetocreate) {
	int[] masks = new int[2 * sizetocreate * sizetocreate];
	for (int i = 1; i<masks.length; i+=2) { //i = 1 so that odd masks are turned on (patch transparent)
	    masks[i] = ON;
	}
	//add a turtle transparent square of 4 pixels so shapes aren't accidently lost
	//System.out.println("sizetocreate = " + sizetocreate);
	masks[sizetocreate*sizetocreate - sizetocreate - 2] = ON;
	masks[sizetocreate*sizetocreate - sizetocreate - 1] = OFF;
	masks[sizetocreate*sizetocreate - sizetocreate] = ON;
	masks[sizetocreate*sizetocreate - sizetocreate + 1] = OFF;
	masks[sizetocreate*sizetocreate + sizetocreate - 2] = ON;
	masks[sizetocreate*sizetocreate + sizetocreate - 1] = OFF;
	masks[sizetocreate*sizetocreate + sizetocreate] = ON;
	masks[sizetocreate*sizetocreate + sizetocreate + 1] = OFF;
		
	return masks;
    }

    static void makeTurtleCase(double[] colors, int[] masks) {
	for (int i=0; i<MEDIUM*MEDIUM; i++) {
	    switch (i) {
	    case 23: case 24:
	    case 38: case 39: case 40: case 41:
	    case 54: case 55: case 56: case 57:
	    case 67: case 68: case 70: case 71: case 72: case 73: case 75: case 76:
	    case 82: case 83: case 84: case 85: case 86: case 87:
	    case 88: case 89: case 90: case 91: case 92: case 93:

	    case 99: case 100: case 101: case 102: case 103:
	    case 104: case 105: case 106: case 107: case 108:

	    case 116: case 117: case 118: case 119:
	    case 120: case 121: case 122: case 123:

	    case 132: case 133: case 134: case 135:
	    case 136: case 137: case 138: case 139:

	    case 148: case 149: case 150: case 151:
	    case 152: case 153: case 154: case 155:

	    case 164: case 165: case 166: case 167:
	    case 168: case 169: case 170: case 171:

	    case 180: case 181: case 182: case 183:
	    case 184: case 185: case 186: case 187:

	    case 195: case 196: case 197: case 198: case 199:
	    case 200: case 201: case 202: case 203: case 204:

	    case 210: case 211: case 212: case 213: case 214: case 215:
	    case 216: case 217: case 218: case 219: case 220: case 221:

	    case 227: case 228: case 230: case 231:
	    case 232: case 233: case 235: case 236:
		masks[2*i]   = ON;
		masks[2*i+1] = OFF;
		break;
	    default:
		masks[2*i]   = OFF;
		masks[2*i+1] = ON;
	    }
	}
    }



    static void makePersonCase(double[] colors, int[] masks) {

	for (int i = 0; i < MEDIUM*MEDIUM; i++) {
	    switch (i) {
	    case 6: case 7: case 8: case 9:
	    case 22: case 23: case 24: case 25:
	    case 32: case 33: case 38: case 39: case 40: case 41: case 46: case 47:
	    case 48: case 49: case 54: case 55: case 56: case 57: case 62: case 63:
	    case 64: case 65: case 66: case 67: case 68: case 69: case 70: case 71:
	    case 72: case 73: case 74: case 75: case 76: case 77: case 78: case 79:
	    case 80: case 81: case 82: case 83: case 84: case 85: case 86: case 87:
	    case 88: case 89: case 90: case 91: case 92: case 93: case 94: case 95:
	    case 102: case 103: case 104: case 105:
	    case 118: case 119: case 120: case 121:
	    case 134: case 135: case 136: case 137:
	    case 150: case 151: case 152: case 153:
	    case 164: case 165: case 166: case 167: case 168: case 169: case 170: case 171:
	    case 180: case 181: case 182: case 183: case 184: case 185: case 186: case 187:
	    case 196: case 197: case 202: case 203:
	    case 212: case 213: case 218: case 219:
	    case 226: case 227: case 228: case 229: case 234: case 235: case 236: case 237:
	    case 242: case 243: case 244: case 245: case 250: case 251: case 252: case 253:
		
		masks[2*i] = ON;
		masks[2*i+1] = OFF;
		break;
	    default:
		masks[2*i] = OFF;
		masks[2*i+1] = ON;
	    }
	}
    }



    static int convert(double coord, boolean x, int size) {
	// abegel: what is the 22 for?
	//coord is x or y; if boolean x = true, add 22, else don't, size is size of colors
	//rather than rescale to fit into these roundings, just cut off.
	int z = (int)Math.round(coord);
	if (size == 8)
	    { if (x) { z = (int)Math.round(z + 22); }
	    else { z = (int)Math.round(z); }
	    }
	if (size == 16)
	    { if (x) { z = (int)Math.round(z + 22); }
	    else { z = (int)Math.round(z); }
	    }
	if (size == 32)
	    { if (x) { z = (int)Math.round(1 * (z + 22)); }
	    else { z = (int)Math.round(1 * z); }
	    }
	return z;
    }

    static void rotate(int heading, double[] src_colors, int[] src_masks,
		       double[] colors, int[] masks) {
	int size = (int)Math.sqrt(src_colors.length);
	int i, boundaries, ax;
	int[] sampleMasks;
	double[] sampleColors;
	double tx, ty, px, py, sintheta, costheta;
	double[] p, c;
	if (heading <= 0 || heading >= 16) {
		System.arraycopy(src_colors, 0, colors, 0, src_colors.length);
		System.arraycopy(src_masks, 0, masks, 0, src_masks.length);
	}

    // trap simple cases here to avoid color conversion ickyness
    if (heading == 0) { // no rotation to perform
        return;
    } else if (heading == 4) { // 90 degrees
        int x, xprime;
        for (int m = 0; m < size; m++) {
            for (int n = 0; n < size; n++) {
                x = size * m + n;
                xprime = size * n + (size - m - 1);
                colors[xprime] = src_colors[x];
                masks[2 * xprime] = src_masks[2 * x];
                masks[2 * xprime + 1] = src_masks[2 * x + 1];
            }
        }
        return;
    } else if (heading == 8) { // 180
        int x, xprime;
        for (int m = 0; m < size; m++) {
            for (int n = 0; n < size; n++) {
                x = size * (m + 1) - (n + 1);
                xprime = size * (size - m - 1) + n;
                colors[xprime] = src_colors[x];
                masks[2 * xprime] = src_masks[2 * x];
                masks[2 * xprime + 1] = src_masks[2 * x + 1];
            }
        }
        return;
    } else if (heading == 12) { // 270
        int x, xprime;
        for (int m = 0; m < size; m++) {
            for (int n = 0; n < size; n++) {
                x = size * m + n;
                xprime = size * (size - n - 1) + m;
                colors[xprime] = src_colors[x];
                masks[2 * xprime] = src_masks[2 * x];
                masks[2 * xprime + 1] = src_masks[2 * x + 1];
            }
        }
        return;
    }

	p = new double[2];
	c = new double[3];
	/*masks = new int[src_masks.length];
	  colors = new double[src_colors.length];*/  //oops - those were already set up
	sampleMasks = new int[2 * 4];
	sampleColors = new double[4];
	sintheta = Math.sin(-.3927 * (double)heading);   //.3927 is (pi / 8); negative to reverse the rotation
	costheta = Math.cos(-.3927 * (double)heading);

    // the following cases should no longer be hit; trapped above
	//if(heading == 4) {sintheta = -1; costheta = 0;}
	//else if(heading == 8) {sintheta = 0; costheta = -1;}
	//else if(heading == 12) {sintheta = 1; costheta = 0;}

	ax = 0;
	tx = ty = 0.;
	for(int y = 1 - size; y <= size - 1; y += 2) {
	    for(int x = 1 - size; x <= size - 1; x += 2, ax++) { 
		tx = ((double)x * costheta - (double)y * sintheta) / 2;  //source location
		ty = ((double)y * costheta + (double)x * sintheta) / 2;
		px = tx - (int)tx + (tx < 0 ? 1 : 0); //(x, y) location relative to top left of pixel mapped to
		py = ty - (int)ty + (ty < 0 ? 1 : 0);
		px = px - .5; // (x,y) location of top left of sampling square, relative to source pixel top left
		py = py - .5;
		tx = (int)tx - (tx < 0 ? 1 : 0) + size / 2; //x, y location of top left pixel of sampling square, colors[] coordinates
		ty = (int)ty - (ty < 0 ? 1 : 0) + size / 2;
		//adjust to correctly center, and make px, py positive
		if(px < 0) {tx -= 1; px = -px;} else {px = 1 - px;}           
		if(py < 0) {ty -= 1; py = -py;} else {py = 1 - py;}
		boundaries = (ty < 0 ? 1 : 0) + (ty >= size - 1 ? 2 : 0) + (tx < 0 ? 4 : 0) + (tx >= size - 1 ? 8 : 0); //for boundary condition handling
		if((ty < -1) || (ty >= size) || (tx < -1) || (tx >= size))  { //completely out of bounds
		    for(i = 0; i < 4; i++) {
			sampleMasks[2 * i] = OFF;
			sampleMasks[2 * i + 1] = ON;
			sampleColors[i] = 0;
		    }
		}
		else {                                                //at least one patch of the sample is in bounds
		    
		    for(i = 0; i < 2; i++) {
			sampleMasks[0 + i] = ((boundaries & 5)  != 0 ? (i == 1 ? ON : OFF) : src_masks[((int)tx + size * (int)ty) * 2 + i]);
			sampleMasks[2 + i] = ((boundaries & 9)  != 0 ? (i == 1 ? ON : OFF) : src_masks[((int)tx + 1 + size * (int)ty) * 2 + i]);
			sampleMasks[4 + i] = ((boundaries & 6)  != 0 ? (i == 1 ? ON : OFF) : src_masks[((int)tx + size * ((int)ty + 1)) * 2 + i]);
			sampleMasks[6 + i] = ((boundaries & 10) != 0 ? (i == 1 ? ON : OFF) : src_masks[((int)tx + 1 + size * ((int)ty + 1)) * 2 + i]);
		    }
		    sampleColors[0] = ((boundaries & 5)  != 0 || sampleMasks[0] == ON || sampleMasks[1] == ON ? 0. : src_colors[(int)tx + size * (int)ty]);
		    sampleColors[1] = ((boundaries & 9)  != 0 || sampleMasks[2] == ON || sampleMasks[3] == ON ? 0. : src_colors[(int)tx + 1 + size * (int)ty]);
		    sampleColors[2] = ((boundaries & 6)  != 0 || sampleMasks[4] == ON || sampleMasks[5] == ON ? 0. : src_colors[(int)tx + size * ((int)ty + 1)]);
		    sampleColors[3] = ((boundaries & 10) != 0 || sampleMasks[6] == ON || sampleMasks[7] == ON ? 0. : src_colors[(int)tx + 1 + size * ((int)ty + 1)]);
		}
		for(i = 0; i < 2; i++) {
		    p[i] = px * py * (double)sampleMasks[0 + i] +
			(1 - px) * py * (double)sampleMasks[2 + i] +
			px * (1 - py) * (double)sampleMasks[4 + i] +
			(1 - px) * (1 - py) * (double)sampleMasks[6 + i];
		}
		if(p[0] <= p[1] && p[0] <= -1 - p[0] - p[1]) {
		    masks[ax * 2 + 0] = ON;
		    masks[ax * 2 + 1] = OFF;
		    colors[ax] = 0.;
		}
		else if(p[1] < p[0] && p[1] <= -1 - p[0] - p[1]) {
		    masks[ax * 2 + 0] = OFF;
		    masks[ax * 2 + 1] = ON;
		    colors[ax] = 0.;
		}
		else {
		    masks[ax * 2 + 0] = OFF;
		    masks[ax * 2 + 1] = OFF;
		    for(i = 0; i < 3; i++)  {
			c[i] = px * py * (double)((Colors.colorarrayints[(int)(sampleColors[0] * 32.0)] >> (8 * i)) & 0xFF);
			c[i] += (1 - px) * py * (double)((Colors.colorarrayints[(int)(sampleColors[1] * 32.0)] >> (8 * i)) & 0xFF);
			c[i] += px * (1 - py) * (double)((Colors.colorarrayints[(int)(sampleColors[2] * 32.0)] >> (8 * i)) & 0xFF);
			c[i] += (1 - px) * (1 - py) * (double)((Colors.colorarrayints[(int)(sampleColors[3] * 32.0)] >> (8 * i)) & 0xFF);
			c[i] /= (1 + p[0] + p[1]);
		    }
		    colors[ax] = Colors.mapColorToStarLogoColor(((int)(c[2] + .5) << 16) + ((int)(c[1] + .5) << 8) + ((int)(c[0] + .5)));
		}
	    }
	}
    }   

    static void flipX(double[] src_colors, int[] src_masks, 
		      double[] dst_colors, int[] dst_masks) {
	int size = (int)Math.sqrt(src_colors.length);
	for (int row=0; row < size; row++) {
	    for (int col=0; col < size; col++) {
		int sindex = col + row*size;
		int findex = col + (size-row-1)*size;
		dst_colors[findex]    = src_colors[sindex];
		dst_masks[2*findex]   = src_masks[2*sindex];
		dst_masks[2*findex+1] = src_masks[2*sindex+1];
	    }
	}
    }

    static void flipY(double[] src_colors, int[] src_masks, 
		      double[] dst_colors, int[] dst_masks) {
	int size = (int)Math.sqrt(src_colors.length);
	for (int row=0; row < size; row++) {
	    for (int col=0; col < size; col++) {
		int sindex = col + row*size;
		int findex = (size-col-1) + row*size;
		dst_colors[findex]    = src_colors[sindex];
		dst_masks[2*findex]   = src_masks[2*sindex];
		dst_masks[2*findex+1] = src_masks[2*sindex+1];
	    }
	}
    }


    static void resize(double[] src_c, int[] src_m, int newSize,
		       double[] colors, int[] masks, Application app) {
			
	//maybe need to change to do the nearest neighbors based on turtle transparency rather than color?

	int size = (int)Math.sqrt(src_c.length);
	int index = 0;
	if (newSize > size) {
	    //System.out.println("resize to bigger size shape");
	    int factor = newSize / size;
	    for (int j=0; j<src_c.length; j++) {
		int row = j/size;
		int column = (j%size);
		for (int k=0; k<factor; k++) {
		    for (int l=0; l<factor; l++) {
			index = (k+row*factor)*newSize + (column*factor+l);
			colors[index]    = src_c[j];
			masks[2*index]   = src_m[2*j];
			masks[2*index+1] = src_m[2*j+1];
		    }
		}
	    }
	} else {
	    //System.out.println("resize to smaller shape");
	    int factor = size / newSize;
	    int z = 0;
	    for (int i=0; i < src_c.length; ) {
		// get all of the important info from the block
		int tt = 0; // number of turtle transparent cells in the block
		int pt = 0; // number of patch transparent cells in the block
		int cl = 0; // number of cells with color in the block
			
		double representativeColor = 15 * 32.0;
			
		for (int j=0; j<factor; j++) {
		    for (int k=0; k<factor; k++) {
			index = i+j*size+k;
			// turtle transparent int
			try {
			    int t = src_m[2*index];
			    if (t != OFF) { tt++; continue; }
			} catch (ArrayIndexOutOfBoundsException e) {
			    app.err.println(src_m.length + " " + (2*index));
			}
			// patch transparent int
			int p = src_m[2*index+1];
			if (p != OFF) { pt++; continue; }
			// color
			double c = src_c[index];
			if (c != OFF) { cl++; representativeColor = c; continue; }
			colors[j*factor+k] = c;		
		    }
		}

		// using nearest-neighbors, choose a representative color for cells[]
		if (tt >= factor*factor/4) { // if half or more cells are turtle transparen
		    masks[2*z]   = ON;
		    masks[2*z+1] = OFF;
		    colors[z]     = OFF;
		} else if ( cl >= pt ) { // color the turtle
		    masks[2*z]   = OFF;
		    masks[2*z+1] = OFF;
		    colors[z]     = representativeColor;
		} else { // make the cell patch transparent
		    masks[2*z]   = OFF;
		    masks[2*z+1] = ON;
		    colors[z]    = OFF;
		}
		z++;

		// get to the next block
		i += factor;
		if ((i % (size)) == 0) { // if you have reached the end of a line, go to the next one
		    i += size*(factor-1);
		}
	    }
		
	}

    }

    public void saveShape(DataOutputStream oos) throws IOException {
	//System.out.println("save shape: " + name + " id: " + id);
	oos.writeUTF(name);
	oos.writeInt(id);
	oos.writeBoolean(rotatep);
	
	oos.writeInt(3); // num sizes
	
	oos.writeInt(SMALL);
	double[] colors = this.colors8x8[0];
	oos.writeInt(colors.length);
	for (int i = 0; i < colors.length; i++) {
	    oos.writeDouble(colors[i]);
	}
	int[] masks = this.masks8x8[0];
	oos.writeInt(masks.length);
	for (int i = 0; i < masks.length; i++) {
	    oos.writeInt(masks[i]);
	}

	oos.writeInt(MEDIUM);
	colors = this.colors16x16[0];
	oos.writeInt(colors.length);
	for (int i = 0; i < colors.length; i++) {
	    oos.writeDouble(colors[i]);
	}
	masks = this.masks16x16[0];
	oos.writeInt(masks.length);
	for (int i = 0; i < masks.length; i++) {
	    oos.writeInt(masks[i]);
	}

	oos.writeInt(LARGE);
	colors = this.colors32x32[0];
	oos.writeInt(colors.length);
	for (int i = 0; i < colors.length; i++) {
	    oos.writeDouble(colors[i]);
	}
	masks = this.masks32x32[0];
	oos.writeInt(masks.length);
	for (int i = 0; i < masks.length; i++) {
	    oos.writeInt(masks[i]);
	}

    }
    
    public static TurtleShapes restoreShape(DataInputStream ois, double version, Application app) throws IOException {
	//	System.out.println("restoring shape: " + version);
	String name = "";
	try {
	    name = ois.readUTF();
	} catch (UTFDataFormatException e) {
	    app.err.println("restore shape failed");
	    return null;
	}
	//System.out.println("name is " + name); 
	int id = ois.readInt();
	//System.out.println("id is " + id);
	boolean rotatep = false;
	if (version == 2.0) {
	    rotatep = (ois.readInt() == 0) ? false : true;
	} else if (version == 3.0) {
	    rotatep = ois.readBoolean();
	    //System.out.println("read rotatep: " + rotatep);
	}
	if (version == 2.0) {
	    double[] colors = new double[ois.readInt()];
	    for(int i = 0; i < colors.length; i++) {
		colors[i] = ois.readDouble();
	    }
	    int[] masks = new int[ois.readInt()];
	    for (int i = 0; i < masks.length; i++) {
		masks[i] = ois.readInt();
	    }
	    int size = (int)Math.sqrt(colors.length);
	    switch (size) {
	    case SMALL:
		return new TurtleShapes(name, id, colors, masks, null, null, null, null, rotatep);
	    case MEDIUM:
		return new TurtleShapes(name, id, null, null, colors, masks, null, null, rotatep);
	    case LARGE:
		return new TurtleShapes(name, id, null, null, null, null, colors, masks, rotatep);
	    }
	    return null;
	} else if (version == 3.0) {
	    int numSizes = ois.readInt();
	    //System.out.println("sizes: " + numSizes);
	    double[][] colorsArray = new double[numSizes][];
	    int[][] masksArray = new int[numSizes][];
	    for(int i = 0; i < numSizes; i++) {
		int size = ois.readInt();
		//System.out.println("shape size = " + size);
		double[] colors = new double[ois.readInt()];
		//System.out.println("colors array is " + colors.length);
		for(int j = 0; j < colors.length; j++) {
		    colors[j] = ois.readDouble();
		    assert(colors[j] >= 0.0 && colors[j] < (double)Colors.numcolors) : "Illegal color in shape";
		}
		int[] masks = new int[ois.readInt()];
		//System.out.println("masks array is " + masks.length);
		for (int j = 0; j < masks.length; j++) {
		    masks[j] = ois.readInt();
		}
		switch(size) {
		case SMALL:
		    colorsArray[0] = colors;
		    masksArray[0] = masks;
		    break;
		case MEDIUM:
		    colorsArray[1] = colors;
		    masksArray[1] = masks;
		    break;
		case LARGE:
		    colorsArray[2] = colors;
		    masksArray[2] = masks;
		    break;
		}
	    }
	    return new TurtleShapes(name, id, colorsArray[0], masksArray[0], 
				    colorsArray[1], masksArray[1],
				    colorsArray[2], masksArray[2],
				    rotatep);
	}
	return null;
    }

    static String saveShapes() {
    	ByteArrayOutputStream shapes;
    	byte[] shapescomp, shapesbytes;
    	String shapeschars;

    	try {
    	    DataOutputStream oos = new DataOutputStream(shapes = new ByteArrayOutputStream());
    	    saveShapes(oos);
    	    oos.close();
    	    shapesbytes = shapes.toByteArray();
    	    shapescomp = Utilities.gzipByteArray(shapesbytes);
    	    shapeschars = yoyo.Base64.encode(shapescomp);
    	    return "`turtle shapes`\r\n" + shapeschars;
    	} catch (IOException e) {
    	    e.printStackTrace();
    	}
    	catch (RuntimeException g) {
    	    g.printStackTrace();
    	}
    	// return an empty set of shapes if you can't save the world
    	return "`turtle shapes`\r\n";
    }

    static String saveShapes(int shapeNum) {
    	ByteArrayOutputStream shapes;
    	byte[] shapescomp, shapesbytes;
    	String shapeschars;

    	try {
    	    DataOutputStream oos = new DataOutputStream(shapes = new ByteArrayOutputStream());
    	    saveShapes(oos, shapeNum);
    	    oos.close();
    	    shapesbytes = shapes.toByteArray();
    	    shapescomp = Utilities.gzipByteArray(shapesbytes);
    	    shapeschars = yoyo.Base64.encode(shapescomp);
    	    return "`turtle shapes`\r\n" + shapeschars;
    	} catch (IOException e) {
    	    e.printStackTrace();
    	}
    	catch (RuntimeException g) {
    	    g.printStackTrace();
    	}
    	// return an empty set of shapes if you can't save the world
    	return "`turtle shapes`\r\n";
    }

    static void saveShapes(DataOutputStream oos) throws IOException {
	// write out this info:
	// version of saveShapes
	// number of shapes for each size, 8x8, 16x16 and 32x32.
	// each num is followed by the list of shapes for that size.
	oos.writeUTF("`turtle shapes`");
	oos.writeDouble(3.0);

	int num = calculateNumShapes();
	oos.writeInt(num);
	for(int i = 0; i < shapes.length; i++) {
	    if (shapes[i] != null) {
		oos.writeInt(i);
		shapes[i].saveShape(oos);
	    }
	}
    }

    static void saveShapes(DataOutputStream oos, int shapeNum) throws IOException {
	// write out this info:
	// version of saveShapes
	// number of shapes for each size, 8x8, 16x16 and 32x32.
	// each num is followed by the list of shapes for that size.
	oos.writeUTF("`turtle shapes`");
	oos.writeDouble(3.0);

	if (shapes[shapeNum] != null) {
	    int num = 1;
	    oos.writeInt(num);
	    oos.writeInt(shapeNum);
	    shapes[shapeNum].saveShape(oos);
	} else {
	    oos.writeInt(0);
	}
    }

    static boolean loadShapes(String str, boolean deleteShapes, Application app) {
	String title = "`turtle shapes`";
	int start = str.indexOf(title);
	if (start >= 0) {
	    int end = str.indexOf("`", start+title.length());
	    if (end == -1) end = str.length();
	    String shapestring = str.substring(start+title.length(), end).trim()+"\r\n";
	    try {
		byte[] bytes = Utilities.gunzipByteArray(yoyo.Base64.decode(shapestring.getBytes()));
		try {
		    DataInputStream ois = new DataInputStream(new ByteArrayInputStream(bytes));
		    restoreShapes(ois, -1, deleteShapes, app);
		} catch (StreamCorruptedException sce) {
		    sce.printStackTrace();
		    loadDefaultShapes(app);
		} catch (IOException e) { 
		    e.printStackTrace(); 
		    loadDefaultShapes(app);
		} catch (AssertionError e) { 
		    e.printStackTrace();
		    loadDefaultShapes(app);
		}
	    } catch (IOException e) {
		e.printStackTrace();
		loadDefaultShapes(app);
	    }
	    finally {
		if (app.shapeChooser != null) app.shapeChooser.refresh();
	    }
	    return true;
	}
	return false;
    }

    static boolean loadShapes(String str, int shapeNum, boolean deleteShapes, Application app) {
	String title = "`turtle shapes`";
	int start = str.indexOf(title);
	if (start >= 0) {
	    int end = str.indexOf("`", start+title.length());
	    if (end == -1) end = str.length();
	    String shapestring = str.substring(start+title.length(), end).trim()+"\r\n";
	    try {
		byte[] bytes = Utilities.gunzipByteArray(yoyo.Base64.decode(shapestring.getBytes()));
		try {
		    DataInputStream ois = new DataInputStream(new ByteArrayInputStream(bytes));
		    restoreShapes(ois, shapeNum, deleteShapes, app);
		} catch (StreamCorruptedException sce) {
		    sce.printStackTrace();
		    loadDefaultShapes(app);
		} catch (IOException e) { 
		    e.printStackTrace(); 
		    loadDefaultShapes(app);
		} catch (AssertionError e) { 
		    e.printStackTrace();
		    loadDefaultShapes(app);
		}
	    } catch (IOException e) {
		e.printStackTrace();
		loadDefaultShapes(app);
	    } 
	    finally {
		if (app.shapeChooser != null) app.shapeChooser.refresh();
	    }
	    return true;
	}
	return false;
    }

    static void restoreShapes(DataInputStream ois, int shapeNum, boolean deleteShapes, Application app) 
	throws IOException {
	String header = "";
	try {
	    header = ois.readUTF();
	}
	catch (UTFDataFormatException e) {
	    return;
	}
	if (!header.equals("`turtle shapes`")) {
	    //System.out.println("wrong header in restore shapes.  returning");
	    return;
	}

	double version = ois.readDouble();
	if (version == 2.0) {
	    // kill all shapes here
	    if (deleteShapes) shapes = new TurtleShapes[MAXSHAPES];
	    
	    // read 8x8 shapes
	    int num = ois.readInt();
	    for (int i = 0; i < num; i++) {
		int index = ois.readInt();
		//System.out.println("restoring 8x8");
		shapes[(shapeNum < 0) ? index : shapeNum] = restoreShape(ois, version, app);
	    }
	    
	    // read 16x16 shapes
	    num = ois.readInt();
	    for (int i = 0; i < num; i++) {
		int index = ois.readInt();
		TurtleShapes loadingShape = restoreShape(ois, version, app);
		TurtleShapes shape = shapes[(shapeNum < 0) ? index : shapeNum];
		if (shape == null) {
		    shapes[(shapeNum < 0) ? index : shapeNum] = loadingShape;
		} else {
		    shape.colors16x16 = loadingShape.colors16x16;
		    shape.masks16x16 = loadingShape.masks16x16;
		}
	    }
	    
	    // read 32x32 shapes
	    num = ois.readInt();
	    for (int i = 0; i < num; i++) {
		int index = ois.readInt();
		TurtleShapes loadingShape = restoreShape(ois, version, app);
		TurtleShapes shape = shapes[(shapeNum < 0) ? index : shapeNum];
		if (shapes == null) {
		    shapes[(shapeNum < 0) ? index : shapeNum] = loadingShape;
		} else {
		    shape.colors32x32 = loadingShape.colors32x32;
		    shape.masks32x32 = loadingShape.masks32x32;
		}
	    }

	} else if (version > 3.0 || version < 3.0) {
	    StarLogoError.error(app, app.textBundle.getString("error_loading_project") +
				": " + app.textBundle.getString("error_cant_read_shapes") +
				". " + app.textBundle.getString("error_version_unknown") +
				": " + version, false);
	    //System.out.println("wrong version, returning");
	    return;
	} else {
	    // kill all shapes here
	    if (deleteShapes) shapes = new TurtleShapes[MAXSHAPES];

	    int num = ois.readInt();
	    for (int i = 0; i < num; i++) {
		int index = ois.readInt();
		//System.out.println("restoring shape #" +i);
		shapes[(shapeNum < 0) ? index : shapeNum] = restoreShape(ois, version, app);
	    }
	}

    }
    
    private static class ShapeClipboard {
	TurtleShapes shape;
	boolean cutp;

	ShapeClipboard(int num, boolean cutp) {
	    shape  = TurtleShapes.getShape(num).copy();
	    this.cutp = cutp;
	}
    }

    static {
	try {
	    new Thread(new Runnable() {
		    // Cleanup thread. Runs every 60 seconds or so to clean out the pools.
		    public void run() {
			Thread thread = Thread.currentThread();
			while (true) {
			    try {
				thread.sleep(60000);
			    }
			    catch (InterruptedException e) {}
			    clearHash();
			}
		    }
		}).start();
	} catch (Exception e) { e.printStackTrace(); }
    }

}
